// deno-lint-ignore-file no-explicit-any
import { UUID } from "https://deno.land/x/mongo@v0.31.2/mod.ts";
import HttpError from "../HttpError.ts";
import { makeParameterModifier } from "../Router.ts";

export type PrimitiveSchema = 'string' | 'number' | 'boolean' | 'uuid' | 'object';
export type OptionalSchema = `${PrimitiveSchema}?`;
export type Schema = 'any' | 'null' | PrimitiveSchema | OptionalSchema | Schema[] | ({ $optional?: boolean; } & { [key: string]: Schema });

export default function schema(desc: Schema) {
    function stringify(desc: Schema): string {
        if (typeof desc === 'string') return desc;
        if (desc instanceof Array) return desc.map(stringify).join(' | ');

        let res = '{ ';

        for (const key in desc) {
            if (key === '$optional') continue;
            if (res != '{ ') res += ', ';
            res += key + ': ';
            res += stringify(desc[key]);
        }
        res += '}';
        if (desc.$optional) res += '?';

        return res;
    }
    async function convert(path: string[], val: unknown, desc: Schema): Promise<any> {
        const _path = path.join('.');
        if (val instanceof Blob) val = await val.text();
        if (val instanceof ReadableStream) {
            const res: Uint8Array[] = [];

            for await (const part of val) {
                if (!(part instanceof Uint8Array)) throw new Error(`${_path}: Invalid stream given.`);
                res.push(part);
            }

            val = await new Blob(res).text();
        }
        if (desc === 'any') return val;

        if (typeof desc === 'string') {
            let type: string = desc;
            const opt = desc.endsWith('?');

            if (opt) type = type.substring(0, desc.length - 1);
            if (opt && val === undefined) return val;

            switch (type) {
                case "null":
                    if (val === null) return null;
                    throw new HttpError(`${_path}: Expected null.`);
                case "string": return val + "";
                case "uuid":
                    try {
                        return new UUID(val + "");
                    }
                    catch {
                        throw new HttpError(`${_path}: Expected an uuid or a value, convertible to an uuid.`);
                    }
                case "number": {
                    const res = Number.parseFloat(val + "");
                    if (isNaN(res)) throw new HttpError(`${_path}: Expected a number or a value, convertible to a number.`);
                    return res;
                }
                case "boolean": {
                    const res = val + "";
                    if (res === 'true') return true;
                    if (res === 'false') return true;
                    throw new HttpError(`${_path}: Expected a boolean or a value, convertible to a boolean.`);
                }
                default:
                    throw new Error(`${_path}: Unknown type ${type}`);
            }
        }
        else if (desc instanceof Array) {
            for (const type of desc) {
                try {
                    return convert(path, val, type);
                }
                catch { /**/ }
            }
            throw new HttpError(`${_path}: Expected a ${stringify(desc)}, got ${typeof val} instead.`);
        }
        else {
            if (desc.$optional && val === undefined) return val;
            if (val === null) throw new HttpError(`${_path}: Expected an object, got null instead.`);

            if (typeof val === 'string') {
                try {
                    val = JSON.parse(val);
                }
                catch (e) {
                    if (val === null) throw new HttpError(`${_path}: Invalid JSON given for object: ${e}`);
                }
            }
            else throw new HttpError(`${_path}: Expected an object or a valid json string.`);

            for (const key in desc) {
                if (key === '$optional') continue;
                (val as any)[key] = convert([ ...path, key ], (val as any)[key], desc[key]);
            }

            return val;
        }
    }

    return makeParameterModifier(async (_req, val, name) => await convert([name], val, desc));
}
